@@ -122,9 +122,10 @@ void WinExit(void);
 static void WinCloseHelp(void);
 int CALLBACK ShutDown(void);
 #ifdef WGP_CONSOLE
-static char * ConsoleGetS(char * str, unsigned int size);
+static char * ConsoleGetS(char * str, unsigned int size, TBOOLEAN echo);
 static int ConsolePutS(const char *str);
 static int ConsolePutCh(int ch);
+static size_t read_stdin_pipe(void *ptr, size_t len);
 #endif
 
 
@@ -796,7 +797,13 @@ MyFGetC(FILE *file)
 {
     if (isterm(file))
 	return GETCH();
+#ifndef WGP_CONSOLE
     return fgetc(file);
+#else
+    if (file != stdin)
+	return fgetc(file);
+    return ConsoleGetch();
+#endif
 }
 
 char *
@@ -811,17 +818,19 @@ MyGetS(char *str)
 char *
 MyFGetS(char *str, unsigned int size, FILE *file)
 {
-    if (isterm(file)) {
 #ifndef WGP_CONSOLE
+    if (isterm(file)) {
 	char * p = TextGetS(&textwin, str, size);
 	if (p != NULL)
 	    return str;
 	return NULL;
-#else
-	return ConsoleGetS(str, size);
-#endif
     }
     return fgets(str,size,file);
+#else
+    if (file != stdin)
+	return fgets(str, size, file);
+    return ConsoleGetS(str, size, isterm(file));
+#endif
 }
 
 int
@@ -950,7 +959,13 @@ MyFRead(void *ptr, size_t size, size_t n, FILE *file)
 	TEXTMESSAGE;
 	return n;
     }
+#ifndef WGP_CONSOLE
     return fread(ptr, size, n, file);
+#else
+    if (file != stdin)
+	return fread(ptr, size, n, file);
+    return read_stdin_pipe(ptr, size * n) / size;
+#endif
 }
 
 
@@ -1083,19 +1098,27 @@ fake_pclose(FILE *stream)
 HANDLE input_thread = NULL;
 HANDLE input_event = NULL;
 HANDLE input_cont = NULL;
-int nextchar = EOF;
+#define READBUFSIZE 1024
+static char readbuf[2][READBUFSIZE];
+static int rbuf_page = 0;
+static char *rbuf = NULL;
+static size_t rbuf_len = 0;
+static size_t rbuf_idx = 0;
+static int pipe_eof = 0;
 
 DWORD WINAPI
 stdin_pipe_reader(LPVOID param)
 {
     do {
-	unsigned char c;
-	size_t n = fread(&c, 1, 1, stdin);
+	size_t n = read(_fileno(stdin), readbuf[rbuf_page], READBUFSIZE);
+	if (n < 0) /* This should not happen. */
+	    n = 0;
 	WaitForSingleObject(input_cont, INFINITE);
-	if (n == 1)
-	    nextchar = c;
-	else if (feof(stdin))
-	    nextchar = EOF;
+	rbuf = readbuf[rbuf_page];
+	rbuf_page ^= 1;
+	rbuf_len = n;
+	rbuf_idx = 0;
+	pipe_eof = (n == 0);
 	SetEvent(input_event);
 	Sleep(0);
     } while (TRUE);
@@ -1107,7 +1130,7 @@ HANDLE
 init_pipe_input(void)
 {
     if (input_event == NULL)
-	input_event = CreateEvent(NULL, FALSE, FALSE, NULL);
+	input_event = CreateEvent(NULL, TRUE, FALSE, NULL);
     if (input_cont == NULL)
 	input_cont = CreateEvent(NULL, FALSE, TRUE, NULL);
     if (input_thread == NULL)
@@ -1119,8 +1142,14 @@ init_pipe_input(void)
 int
 next_pipe_input(void)
 {
-    int c = nextchar;
-    SetEvent(input_cont);
+    char c;
+    if (pipe_eof)
+	return EOF;
+    c = rbuf[rbuf_idx++];
+    if (rbuf_idx == rbuf_len) {
+	ResetEvent(input_event);
+	SetEvent(input_cont);
+    }
     return c;
 }
 
@@ -1132,11 +1161,10 @@ ConsoleGetch(void)
     HANDLE h;
     DWORD waitResult;
 
-    if (!_isatty(fd)) {
+    if (!_isatty(fd))
 	h = init_pipe_input();
-    } else {
+    else
 	h = (HANDLE)_get_osfhandle(fd);
-    }
 
     do {
 	waitResult = MsgWaitForMultipleObjects(1, &h, FALSE, INFINITE, QS_ALLINPUT);
@@ -1159,6 +1187,32 @@ ConsoleGetch(void)
     return '\r';
 }
 
+
+size_t
+read_stdin_pipe(void *buf, size_t len)
+{
+    size_t cnt = 0;
+
+    while (len > cnt) {
+	size_t copy_len;
+
+	WaitForSingleObject(input_event, INFINITE);
+	if (pipe_eof)
+	    break;
+	copy_len = len - cnt;
+	if (copy_len > rbuf_len - rbuf_idx)
+	    copy_len = rbuf_len - rbuf_idx;
+	memcpy((char *)buf + cnt, rbuf + rbuf_idx, copy_len);
+	rbuf_idx += copy_len;
+	cnt += copy_len;
+	if (rbuf_idx == rbuf_len) {
+	    ResetEvent(input_event);
+	    SetEvent(input_cont);
+	}
+    }
+    return cnt;
+}
+
 #endif /* WGP_CONSOLE */
 
 
@@ -1227,7 +1281,7 @@ ConsoleReadCh(void)
 #ifdef WGP_CONSOLE
 
 static char *
-ConsoleGetS(char * str, unsigned int size)
+ConsoleGetS(char * str, unsigned int size, TBOOLEAN echo)
 {
     char * next = str;
 
@@ -1239,20 +1293,26 @@ ConsoleGetS(char * str, unsigned int size)
 		return NULL;
 	    return str;
 	case '\r':
-	    *next = '\n';
+	    if (!echo)
+		*next = ConsoleGetch();
+	    else
+		*next = '\n';
 	    /* intentionally fall through */
 	case '\n':
-	    ConsolePutCh(*next);
+	    if (echo)
+		ConsolePutCh(*next);
 	    *(next + 1) = 0;
 	    return str;
 	case 0x08:
 	case 0x7f:
-	    ConsolePutCh(*next);
+	    if (echo)
+		ConsolePutCh(*next);
 	    if (next > str)
 		--next;
 	    break;
 	default:
-	    ConsolePutCh(*next);
+	    if (echo)
+		ConsolePutCh(*next);
 	    ++next;
 	}
     }
@@ -1266,7 +1326,7 @@ ConsolePutS(const char *str)
 {
     LPWSTR wstr = UnicodeText(str, encoding);
     // Word-wrapping on Windows 10 (now) works anyway.
-    if (isatty(fileno(stdout))) {
+    if (_isatty(_fileno(stdout))) {
 	// Using standard (wide) file IO screws up UTF-8
 	// output, so use console IO instead. No idea why
 	// it does so, though.
@@ -1291,7 +1351,7 @@ ConsolePutCh(int ch)
     MultiByteAccumulate(ch, w, &count);
     if (count > 0) {
 	w[count] = 0;
-	if (isatty(fileno(stdout))) {
+	if (_isatty(_fileno(stdout))) {
 	    HANDLE h = GetStdHandle(STD_OUTPUT_HANDLE);
 	    WriteConsoleW(h, w, count, NULL, NULL);
 	} else {
